from keraspoj.src.util.tools import generate_uuid


class TikzElement:
    def __init__(self, internal_name, depends_on=None):
        self.internal_name = internal_name
        self.depends_on = depends_on if depends_on is not None else []

    def to_code(self):
        raise NotImplementedError()


class TikzArg(TikzElement):
    def __init__(self, key, value=None):
        self.key = key
        self.value = value
        super().__init__(generate_uuid(), depends_on=None)

    def to_code(self):
        return str(self)

    def __repr__(self):
        return str(self)

    def __str__(self):
        if self.value is None:
            return str(self.key)
        return f"{self.key}={self.value}"


class TikzOptions(TikzElement):
    def __init__(self, *args, **kwargs):
        self.options = [TikzArg(arg) for arg in args]
        self.options.extend([TikzArg(key, value) for key, value in kwargs.items()])
        super().__init__(generate_uuid(), depends_on=None)

    def add_option(self, key, value):
        self.options.append(TikzArg(key, value))

    def add_flag(self, flag):
        self.options.append(TikzArg(flag))

    def to_code(self):
        return str(self)

    def __str__(self):
        return ",".join([str(option) for option in self.options]) if self.options else ""



from typing import Iterable

from keraspoj.src.backend.base import TikzElement
from keraspoj.src.util.config import Config


class Document:
    def __init__(self):
        self.styles = {}
        self.latex_elements: list[TikzElement] = []

    def add_styles(self, styles):
        self.styles = styles

    def add_style(self, style_name, style: TikzElement):
        self.styles[style_name] = style

    @property
    def _styles_code(self):
        out = ""
        for style_name, style in self.styles.items():
            out += f"% style: {style_name}\n"
            out += rf"\tikzstyle{{{style_name}}}=[{style.to_code()}]" + "\n"
        return out

    def generate_header(self):
        x_scale_value = Config.load_float('diagram', 'coordinate-system', 'x', 'value')
        x_scale_unit = Config.load_str('diagram', 'coordinate-system', 'x', 'unit')

        y_scale_value = Config.load_float('diagram', 'coordinate-system', 'y', 'value')
        y_scale_unit = Config.load_str('diagram', 'coordinate-system', 'y', 'unit')

        scale_value = Config.load_float('diagram', 'coordinate-system', 'scale')

        x_scale = f"{x_scale_value}{x_scale_unit}"
        y_scale = f"{y_scale_value}{y_scale_unit}"

        return "\\documentclass[border=1cm]{standalone}\n" \
               "\\usepackage{xcolor}\n" \
               "\\usepackage{tikz}\n" \
               "\\usetikzlibrary{positioning, shapes.multipart, calc, graphs, graphs.standard, arrows.meta}\n" \
               "\\begin{document}\n" \
               F"\\begin{{tikzpicture}}[x={x_scale}, y={y_scale}, scale={scale_value}]\n" \
               f"{self._styles_code}\n"

    def generate_footer(self):
        return r"\end{tikzpicture}" \
               r"\end{document}"

    def add_element(self, element: TikzElement):
        self.latex_elements.append(element)

    def add_elements(self, elements: Iterable[TikzElement]):
        self.latex_elements.extend(elements)

    def generate_code(self):
        return \
                self.generate_header() + \
                "\n".join([element.to_code() for element in self.latex_elements]) + \
                self.generate_footer()



import keras.layers
import networkx as nx

from keraspoj.src.backend.base import TikzOptions
from keraspoj.src.backend.edge import Edge
from keraspoj.src.backend.misc.position import Position
from keraspoj.src.diagram.layers.layer import Layer
from keraspoj.src.diagram.layers.layer_group import LayerGroup

from keraspoj.src.util.config import Config, DotConfig

import numpy as np

from keraspoj.src.util.const import COLOR_MAP
from keraspoj.src.util.tools import pretty_print


def is_trainable_layer(layer: Layer):
    return not is_operation_layer(layer) and not is_utility_layer(layer)


def is_operation_layer(layer: Layer):
    return layer.type_group == 'operation'


def is_utility_layer(layer: Layer):
    return layer.type_group == 'utility'


class DiagramGraph:
    @classmethod
    def create_layer(cls, layer: keras.layers.Layer) -> Layer:
        return Layer(layer)

    def __init__(self, input_keras_layer: keras.layers.Layer, output_keras_layer: keras.layers.Layer, canvas_width=None,
                 canvas_height=None):

        self.edge_colors = COLOR_MAP
        self._input_keras_layer = input_keras_layer
        self._output_keras_layer = output_keras_layer

        self.input_layer = self.create_layer(input_keras_layer)
        self.output_layer = self.create_layer(output_keras_layer)

        self.canvas_width = canvas_width if canvas_width is not None else Config.load_int('canvas', 'width')
        self.canvas_height = canvas_height if canvas_height is not None else Config.load_int('canvas', 'height')

        self.layers = []
        self.layer_groups: list[LayerGroup] = []
        self.layer_groups_map = {}
        self.graph = None
        self.positions = None
        self.layer_map = {}

        self.child_list = {}
        self.parent_list = {}

        self.group_child_list = {}
        self.group_parent_list = {}

        self._preprocess_layers()
        self._build_layer_groups()

        pretty_print(self.layer_groups_map)
        print(self.layers)

        self._preprocess_layer_groups()
        self._build_graph()

    def _preprocess_layers(self):
        self.layers = []

        queue = [self._input_keras_layer]

        while len(queue) > 0:
            layer = queue.pop(0)
            if layer.name not in self.layer_map:
                diagram_layer = self.create_layer(layer)
                self.layers.append(diagram_layer)
                self.layer_map[diagram_layer.name] = diagram_layer
                self.child_list[diagram_layer.name] = [ol.name for ol in diagram_layer.keras_output_layers]
                self.parent_list[diagram_layer.name] = [il.name for il in diagram_layer.keras_input_layers]

                queue.extend(diagram_layer.keras_output_layers)

    def _preprocess_layer_groups(self):
        group_child_list = {}
        group_parent_list = {}

        for layer_group in self.layer_groups:
            group_child_list[layer_group.name] = []
            group_parent_list[layer_group.name] = []
            for layer_name in layer_group.output_layers:
                group_child_list[layer_group.name].append(self.get_group_by_layer_name(layer_name).name)

            for layer_name in layer_group.input_layers:
                group_parent_list[layer_group.name].append(self.get_group_by_layer_name(layer_name).name)

        self.group_child_list = group_child_list
        self.group_parent_list = group_parent_list

    def _build_graph(self):
        self.graph = nx.from_dict_of_lists(self.group_child_list, create_using=nx.DiGraph)
        self.positions = nx.nx_agraph.graphviz_layout(self.graph, prog='dot', args=DotConfig.load_dot_args())

        # scale positions to fit in the canvas
        min_x = min([pos[0] for pos in self.positions.values()])
        min_y = min([pos[1] for pos in self.positions.values()])
        max_x = max([pos[0] for pos in self.positions.values()])
        max_y = max([pos[1] for pos in self.positions.values()])

        for group_name, pos in self.positions.items():
            new_x = (pos[0] - min_x) / (max_x - min_x) * self.canvas_width if max_x != min_x else 0
            new_y = (pos[1] - min_y) / (max_y - min_y) * self.canvas_height if max_y != min_y else 0

            new_x = round(new_x, 2)
            new_y = round(new_y, 2)

            self.positions[group_name] = Position(new_x, new_y)

    def _build_layer_groups(self):
        self.layer_groups = []
        processed_layers = set()
        queue: list[Layer] = [self.input_layer]
        while len(queue) > 0:
            layer = queue.pop(0)

            for child_layer in self.get_child_layers(layer.name):
                if child_layer.name not in processed_layers:
                    queue.append(child_layer)

            if layer.name in processed_layers:
                continue

            children = self.get_child_layers(layer.name)
            parents = self.get_parent_layers(layer.name)

            if is_trainable_layer(layer):

                utility_before = None
                utility_after = None

                if len(children) == 1 and is_utility_layer(children[0]):
                    utility_after = children[0]
                if len(parents) == 1 and is_utility_layer(parents[0]):
                    utility_before = parents[0]

                if utility_before is not None:
                    if utility_before.name in processed_layers:
                        utility_before = None
                    else:
                        processed_layers.add(utility_before.name)

                if utility_after is not None:
                    if utility_after.name in processed_layers:
                        utility_after = None
                    else:
                        processed_layers.add(utility_after.name)

                layer_group = LayerGroup(layer, utility_before, utility_after)
                self.layer_groups.append(layer_group)
                self.layer_groups_map[layer_group.name] = layer_group
            elif not is_utility_layer(layer):
                layer_group = LayerGroup(layer)
                self.layer_groups.append(layer_group)
                self.layer_groups_map[layer_group.name] = layer_group
            else:
                if is_utility_layer(layer):
                    layer_group = LayerGroup(layer)
                    self.layer_groups.append(layer_group)
                    self.layer_groups_map[layer_group.name] = layer_group
                    processed_layers.add(layer.name)
            if not is_utility_layer(layer):
                processed_layers.add(layer.name)

    def get_group_by_layer_name(self, layer_name: str) -> LayerGroup:
        for layer_group in self.layer_groups:
            if layer_group.contains_layer_name(layer_name):
                return layer_group
        raise ValueError(f'Layer {layer_name} not found')

    def get_layer(self, name: str) -> Layer:
        return self.layer_map[name]

    def get_child_layers(self, name: str) -> list[Layer]:
        return [self.get_layer(layer_name) for layer_name in self.child_list[name]]

    def get_parent_layers(self, name: str) -> list[Layer]:
        return [self.get_layer(layer_name) for layer_name in self.parent_list[name]]

    def get_layer_by_type(self, layer_type: str) -> list[Layer]:
        return [layer for layer in self.layers if layer.type == layer_type]

    def create_nodes(self):
        nodes = []
        for layer_group in self.layer_groups:
            layer_position = self.positions[layer_group.name]
            nodes.append(layer_group.create_node_group(layer_position))
        return nodes

    def edge_is_skip_connection(self, parent_layer_name, child_layer_name):
        parent_position = self.positions[parent_layer_name]
        child_position = self.positions[child_layer_name]

        if parent_position.x - 0.5 <= child_position.x <= parent_position.x + 0.5:
            range_rectangle = (
                Position(min(parent_position.x, child_position.x) - 1, min(parent_position.y, child_position.y)),
                Position(max(parent_position.x, child_position.x) + 1, max(parent_position.y, child_position.y)))

            for layer_group, position in self.positions.items():
                if layer_group == parent_layer_name or layer_group == child_layer_name:
                    continue
                if (range_rectangle[0].x <= position.x <= range_rectangle[1].x and
                        range_rectangle[0].y <= position.y <= range_rectangle[1].y):
                    # print(f'edge {parent_layer_name} -> {child_layer_name} is crossing {layer_group}')
                    return True

    def calc_bend_direction(self, parent_layer_name, child_layer_name):
        # get other children / parents of parent / child
        parent_children = self.group_child_list[parent_layer_name]
        child_parents = self.group_parent_list[child_layer_name]

        # filter out the current layer
        parent_children = [layer for layer in parent_children if layer != child_layer_name]
        child_parents = [layer for layer in child_parents if layer != parent_layer_name]

        # combine the two lists
        other_layers = parent_children + child_parents

        # average x position of the other layers
        other_layers_x = [self.positions[layer].x for layer in other_layers]
        if len(other_layers_x) == 0:
            return 'right'
        other_layers_x_avg = sum(other_layers_x) / len(other_layers_x)

        # if the average x position is smaller than the child layer, the bend direction is left
        if other_layers_x_avg < self.positions[child_layer_name].x:
            return 'left'
        else:
            return 'right'

    def calc_extra_edge_args(self, parent_layer_name, child_layer_name):
        options = TikzOptions()
        parent_position = self.positions[parent_layer_name]
        child_position = self.positions[child_layer_name]

        parent_child_list = self.group_child_list[parent_layer_name]
        child_parent_list = self.group_parent_list[child_layer_name]

        parent_child_distances = [parent_position.y_distance(self.positions[layer_name]) for layer_name in
                                  parent_child_list]

        child_parent_distances = [child_position.y_distance(self.positions[layer_name]) for layer_name in
                                  child_parent_list]

        min_parent_child_distance = min(parent_child_distances)
        min_child_parent_distance = min(child_parent_distances)
        # y_distance = round(parent_position.y_distance(child_position), 3)
        x_distance = round(parent_position.x_distance(child_position), 3)
        # distance = round(parent_position.distance(child_position), 3)

        ratio_in = 0.3
        ratio_out = 0.3

        if 0 < x_distance < 1:
            ratio_in = ratio_out = 0.1

        if min_parent_child_distance >= 1.5 * min_child_parent_distance:
            # print(f"edge between {parent_layer_name} and {child_layer_name} is a long edge (out)")
            ratio_out = 0.5
        elif min_child_parent_distance >= 1.5 * min_parent_child_distance:
            # print(f"edge between {parent_layer_name} and {child_layer_name} is a long edge (in)")
            ratio_in = 0.5

        in_distance = round(min_child_parent_distance * ratio_in, 2)
        out_distance = round(min_parent_child_distance * ratio_out, 2)

        if self.edge_is_skip_connection(parent_layer_name, child_layer_name):
            in_distance *= 1.25
            out_distance *= 1.25
            bend_direction = self.calc_bend_direction(parent_layer_name, child_layer_name)

            options.add_option(f'bend {bend_direction}', '70')
            options.add_option('in', '180')

        # print(f"between {parent_layer_name} and {child_layer_name} "
        #       f"(y distance: {y_distance}, x distance: {x_distance}, distance: {distance}) "
        #       f"in_distance: {in_distance} out_distance: {out_distance}")

        options.add_option('in distance', f"{in_distance}cm")
        options.add_option('out distance', f"{out_distance}cm")
        return options

    def create_edges(self, use_color=False):
        amount_of_bins = len(self.edge_colors)
        eligible_trainable_layers = [layer for layer in self.layers if is_trainable_layer(layer)]
        trainable_params_list = [layer.trainable_params for layer in eligible_trainable_layers]
        trainable_params_list = [params for params in trainable_params_list if params != 0]

        trainable_params_list = sorted(trainable_params_list)
        if use_color:
            bins = np.array_split(trainable_params_list, amount_of_bins - 1)
            trainable_params_list_bins = [bin[0] for bin in bins]
            # print(sorted(trainable_params_list))

            # trainable_params_list_bins = np.cumsum(trainable_params_list_bins)
            # print("bins:", trainable_params_list_bins)

        def get_edge_color(edge_weight):
            if not use_color or edge_weight == 0:
                return "black"
            # calculate the bin index of the edge weight
            bin_index = np.digitize(edge_weight, trainable_params_list_bins)
            # print(f"edge weight {edge_weight} is in bin {bin_index}")
            return f"COLOR{bin_index}"

        edges = []
        for layer_group in self.layer_groups:
            for child_layer_group_name in self.group_child_list[layer_group.name]:
                child_group = self.layer_groups_map[child_layer_group_name]
                label = child_group.primary_layer.trainable_params

                color = get_edge_color(label)

                label = f'{label:,}' if label != 0 else ''
                extra_args = self.calc_extra_edge_args(layer_group.name, child_layer_group_name)

                edges.append(
                    Edge(layer_group.bottom_layer_name,
                         child_group.top_layer_name,
                         label=label, color=color, extra_args=extra_args))
        return edges



import inflect
from keraspoj.src.backend.base import TikzElement
from keraspoj.src.backend.misc.position import Position
from keraspoj.src.util.const import FONT_SIZE, INNER_SEP, SCALE_Y

from keraspoj.src.util.tools import INDENT_STR, latexify

DEBUG_OFFSET = 0


class Node(TikzElement):
    inflector = inflect.engine()

    def __init__(self, name, *args, node_style_name="default_node",
                 position: Position = None, depends_on=None):
        self.name = name
        self.description_parts = args

        self.position = Position() if position is None else position
        self.node_style_name = node_style_name

        super().__init__(name, depends_on=depends_on)

        if depends_on is None:
            self.depends_on = []

    @property
    def latex_name(self):
        return latexify(self.name)

    @property
    def height(self):
        return ((float(INNER_SEP + FONT_SIZE) * (len(self.description_parts))) - INNER_SEP * 2) / SCALE_Y

    def draw(self):
        return self.to_code()

    def generate_description_code(self):
        text = ""
        description_parts = [str(part) for part in self.description_parts]
        # description_parts.append(self.internal_name)
        for i, part in enumerate(description_parts):
            part = part.replace("_", r"\_")
            part_number = self.inflector.number_to_words(i + 1)
            text += "\n" + INDENT_STR
            text += fr"\nodepart{{{part_number}}}{{{part}}}"
        return text

    def to_code(self):
        layer_description = self.generate_description_code()
        node_code_comment = f"% node: {self.name}\n"

        code = f""

        node_code = rf"\node[{self.node_style_name}] ({self.latex_name}) {self.position.to_code()}" + \
                    f"\n{INDENT_STR}{{{layer_description}}};"
        node_code += "\n"
        code += node_code_comment + node_code

        return code

    def __str__(self):
        return f"Node: {self.name} at {self.position} with style {self.node_style_name}"



import keras.layers

from keraspoj.src.util.config import LayerConfig
from keraspoj.src.util.tools import str_shape
from keraspoj.src.backend.node import Node
from keraspoj.src.backend.misc.position import Position


class Layer:
    sort_order = 9999

    def get_style_name(self) -> str:
        return f"{self.type_group}_layer_style"

    @classmethod
    def get_keras_output_layers(cls, layer: keras.layers.Layer):
        outbound_layers = []
        for outbound_node in layer.outbound_nodes:
            outbound_layers.append(outbound_node.outbound_layer)

        return outbound_layers

    @classmethod
    def get_keras_input_layers(cls, layer: keras.layers.Layer):
        inbound_layers = []
        for inbound_node in layer.inbound_nodes:
            if isinstance(inbound_node.inbound_layers, list):
                for inbound_layer in inbound_node.inbound_layers:
                    inbound_layers.append(inbound_layer)
            else:
                inbound_layers.append(inbound_node.inbound_layers)

        return inbound_layers

    def __init__(self, layer: keras.layers.Layer):
        self.layer = layer
        self.name: str = layer.name
        self.type: str = layer.__class__.__name__
        self.type_group: str = LayerConfig.get_type_group(self.type)
        self.position = Position()

        self.inbound_layers_names = []
        self.outbound_layers_names = []
        self.dependency_layers = set()

        self._parse_input_layers()
        self._parse_output_layers()

    @property
    def trainable_params(self) -> int:
        trainable_params = 0
        for weight in self.layer.weights:
            if weight.trainable:
                trainable_params += weight.shape.num_elements()
        return trainable_params

    @property
    def keras_input_layers(self):
        return self.get_keras_input_layers(self.layer)

    @property
    def keras_output_layers(self):
        return self.get_keras_output_layers(self.layer)

    @property
    def output_shape(self):
        return str_shape(self.layer.output_shape)

    @property
    def input_shape(self):
        return str_shape(self.layer.input_shape)

    def _parse_input_layers(self):
        if len(self.layer.inbound_nodes) == 0:
            return
        for inbound_node in self.layer.inbound_nodes:
            inbound_layers = inbound_node.inbound_layers
            if isinstance(inbound_layers, list):
                for inbound_layer in inbound_layers:
                    self.inbound_layers_names.append(inbound_layer.name)
            else:
                self.inbound_layers_names.append(inbound_layers.name)

    def _parse_output_layers(self):
        if len(self.layer.outbound_nodes) == 0:
            return
        for outbound_node in self.layer.outbound_nodes:
            outbound_layer = outbound_node.outbound_layer
            self.outbound_layers_names.append(outbound_layer.name)

    def _get_siblings(self):
        siblings = set()
        if len(self.keras_input_layers) != 0:
            for inbound_layer in self.keras_input_layers:
                for sibling in self.get_keras_output_layers(inbound_layer):
                    siblings.add(sibling)

        if len(self.keras_output_layers) != 0:
            for outbound_layer in self.keras_output_layers:
                for sibling in self.get_keras_input_layers(outbound_layer):
                    siblings.add(sibling)
        # print(f"Siblings of {self.name}: {siblings}")
        siblings = list(siblings)
        siblings.sort(key=lambda x: x.name)
        return [sibling.name for sibling in siblings]

    def add_dependency_layer(self, layer):
        if isinstance(layer, Layer):
            self.dependency_layers.add(layer.name)
        else:
            self.dependency_layers.add(layer)

    @property
    def layer_description(self) -> tuple:
        layer_type = self.type
        layer = self.layer

        content_lst = [self.type]
        if self.type == "Bidirectional":
            layer_type = self.layer.layer.__class__.__name__
            layer = self.layer.layer
            content_lst = [f"{self.type}({layer_type})"]

        layer_content_config = LayerConfig.load_layer_content(layer_type)

        for content_config in layer_content_config:
            if content_config is None:
                continue

            content_label = content_config['label']
            content_prop = content_config['property']
            content_type = content_config['type']

            if content_prop == 'output_shape':
                content_value = str_shape(layer.output_shape)
            elif content_prop == 'input_shape':
                content_value = str_shape(layer.input_shape)
            else:
                content_value = layer.get_config()
                for prop in content_prop.split('.'):
                    content_value = content_value[prop]

            if content_type == 'int':
                content_value = format(content_value, ",d")
            elif content_type == 'float':
                content_value = format(content_value, ".2f")
            elif content_type == 'shape':
                content_value = str_shape(content_value)

            content_lst.append(f"{content_label}: {content_value}")

        return tuple(content_lst)

    def create_node(self, position: Position = None):

        return Node(self.name,
                    *self.layer_description,
                    node_style_name=self.get_style_name(),
                    position=position,
                    depends_on=[])

    def __repr__(self):
        return self.__str__()
    def __str__(self):
        return f"Layer(name={self.name}, type={self.type})"

    def __lt__(self, other):
        return self.sort_order < other.sort_order or (self.sort_order == other.sort_order and self.name < other.name)



from keraspoj.src.util.config import Config, StyleConfig

SCALE_Y = float(Config.load('diagram', 'coordinate-system', 'y', 'value'))

FONT_SIZE = 6  # pt
INNER_SEP = int(StyleConfig.load('default-style', 'options', 'inner sep')[:-2])  # remove pt


COLOR_MAP = Config.load_colormap()


import os.path
from functools import lru_cache

import yaml

import seaborn as sns

RESOURCE_PATH = os.path.join(os.path.dirname(__file__), '../..', 'resources')


class Config:
    config = None
    config_file = os.path.join(RESOURCE_PATH, 'config', 'config.yaml')

    @classmethod
    def load(cls, *properties):
        with open(cls.config_file) as f:
            cls.config = yaml.full_load(f)
        value = cls.config
        for prop in properties:
            value = value[prop]
        return value

    @classmethod
    def load_int(cls, *properties):
        return int(cls.load(*properties))

    @classmethod
    def load_float(cls, *properties):
        return float(cls.load(*properties))

    @classmethod
    def load_str(cls, *properties):
        return str(cls.load(*properties))

    @classmethod
    def load_colormap(cls):
        colormap_name = cls.load('diagram', 'color', 'colormap')
        colormap_num_colors = cls.load_int('diagram', 'color', 'divisions')

        colormap = sns.color_palette(colormap_name, colormap_num_colors)
        # convert to 0 - 255
        colormap = [(int(r * 255), int(g * 255), int(b * 255)) for r, g, b in colormap]
        return colormap


class GeneralConfig(Config):
    config_file = os.path.join(RESOURCE_PATH, 'config', 'config.yaml')


class StyleConfig(Config):
    config_file = os.path.join(RESOURCE_PATH, 'config', 'style.yaml')

    @classmethod
    def load_styles(cls):
        from keraspoj.src.backend.misc.style import TikzStyle

        cfg_styles = cls.load('styles')
        styles = {}
        for item in cfg_styles:
            style = item['style']
            style_name = item['name']
            styles[style_name] = TikzStyle(*style['flags'], **style['options'])

        return styles


class LayerConfig(Config):
    config_file = os.path.join(RESOURCE_PATH, 'config', 'layers.yaml')

    @classmethod
    @lru_cache(maxsize=128)
    def load_layer_content(cls, layer_type: str):
        content = cls.load('layer-content')
        for group in content['groups']:
            if layer_type in group['layers']:
                # print(f"found layer {layer_type} in group {group['name']}")
                return group['content']
        return content['default']

    @classmethod
    @lru_cache(maxsize=128)
    def get_type_group(cls, layer_type):
        types = cls.load('layer-types')
        for group in types:
            if layer_type in group['layers']:
                return group['name']
        return 'default'


class DotConfig(Config):
    config_file = os.path.join(RESOURCE_PATH, 'config', 'dot.yaml')

    @classmethod
    def load_dot_args(cls):
        dot_config = cls.load('dot')
        dot_args = ""

        for arg_group, options in dot_config.items():
            arg_prefix = arg_group[0].upper()  # it is dirty but works
            for option, value in options.items():
                dot_args += f"-{arg_prefix}{option}={value} "
        return dot_args



import os

from keras.models import load_model
from wand.image import Image as WImage
from wand.color import Color
from keraspoj.src.backend.diagram import Diagram
from keraspoj.src.backend.document import Document
from keraspoj.src.diagram.diagram_graph import DiagramGraph
from keraspoj.src.util.config import StyleConfig, Config

import tex2pix


def create_pdf(document: Document):
    if not os.path.exists('../../out'):
        os.makedirs('../../out')
    with open('../../generated_graph.tex', 'w') as f:
        f.write(document.generate_code())

        r = tex2pix.Renderer(f, runbibtex=True, extras=['example.bib'])
        # r.verbose = True # be loud to the terminal
        # r.rmtmpdir = False # keep the working dir around, for debugging
        r.mkeps('example.eps')
        r.mkpng('example.png')
    # try:
    #     run_command(
    #         'pdflatex -file-line-error -interaction=nonstopmode -synctex=1 -output-format=pdf -output-directory=out '
    #         '/tmp/model-keraspoj/generated_graph.tex > out/log.log')
    #     print("generated pdf file")
    # except Exception as e:
    #     print(e)
    #     print("Error while generating pdf file")
    # parser = LatexLogParser()
    # with open('out/generated_graph.log') as f:
    #     parser.process(f)
    # print(parser)


def visualize(model, resolution=200, canvas_width=None, canvas_height=None):
    input_layer = model.get_layer(name=model.inputs[0].name)
    output_layer = model.get_layer(index=-1)

    if canvas_width is None:
        canvas_width = Config.load_float('canvas', 'width')
    if canvas_height is None:
        canvas_height = Config.load_float('canvas', 'height')

    diagram_graph = DiagramGraph(input_layer, output_layer, canvas_width=canvas_width, canvas_height=canvas_height)
    document = Document()
    diagram = Diagram()

    document.add_styles(StyleConfig.load_styles())

    for node in diagram_graph.create_nodes():
        diagram.add_node(node)

    diagram.add_edges(diagram_graph.create_edges())

    document.add_element(diagram)

    create_pdf(document)

    img = WImage(filename='../../out/generated_graph.pdf', resolution=resolution)
    img.background_color = Color('white')
    img.alpha_channel = 'remove'
    return img


if __name__ == '__main__':
    _model = load_model('model.h5')
    print(_model.get_config())
    visualize(_model, canvas_width=40, canvas_height=500)



from keraspoj.src.backend.base import TikzElement
from keraspoj.src.backend.edge import Edge
from keraspoj.src.backend.node import Node
from keraspoj.src.util.const import COLOR_MAP


class Diagram(TikzElement):
    def __init__(self):
        self.nodes = []
        self.edges = []
        super().__init__("diagram", depends_on=None)

    def add_node(self, node: Node):
        self.nodes.append(node)

    def add_edge(self, edge: Edge):
        self.edges.append(edge)

    def add_edges(self, edges: list[Edge]):
        self.edges.extend(edges)

    @property
    def elements(self) -> list[TikzElement]:
        return self.nodes + self.edges

    @property
    def element_map(self):
        return {element.internal_name: element for element in self.elements}

    def get_drawing_order(self):

        stack_in = self.elements
        stack_out = []
        waiting_line = []

        old_len_stack_out = -1
        old_len_waiting_line = -1

        while len(stack_in) > 0 or len(waiting_line) > 0:
            if len(stack_in) > 0:
                # print(f"there are currently {len(stack_in)} elements in the stack")
                element = stack_in.pop(0)
                if len(element.depends_on) == 0:
                    # print(f"putting element {element.internal_name} on the stack")
                    stack_out.append(element.internal_name)
                else:
                    dependencies_satisfied = True
                    for dependency in element.depends_on:
                        if dependency not in stack_out:
                            dependencies_satisfied = False
                            break
                    if dependencies_satisfied:
                        stack_out.append(element.internal_name)
                    else:
                        # print(
                        # f"putting element {element.internal_name} on the waiting line, because it depends on {element.depends_on}")
                        waiting_line.append(element)

            for waiting_element in waiting_line:
                dependencies_satisfied = True
                for dependency in waiting_element.depends_on:
                    if dependency not in stack_out:
                        dependencies_satisfied = False

                        break
                if dependencies_satisfied:
                    stack_out.append(waiting_element.internal_name)
                    waiting_line.remove(waiting_element)
            # print(f"there are currently {len(waiting_line)} elements left in the waiting line")

            if old_len_stack_out == len(stack_out) and old_len_waiting_line == len(waiting_line):
                print("no progress was made on the waiting line, aborting")
                break
            old_len_stack_out = len(stack_out)
            old_len_waiting_line = len(waiting_line)
        return stack_out

    def register_color_map(self):
        code = ""

        for i, (r, g, b) in enumerate(COLOR_MAP):
            code += rf"\definecolor{{COLOR{i}}}{{RGB}}{{{r},{g},{b}}}" + "\n"
        return code

    def generate_code(self):
        drawing_order = self.get_drawing_order()
        code = self.register_color_map()

        for element_name in drawing_order:
            code += self.element_map[element_name].to_code() + "\n"

        return code

    def to_code(self):
        return self.generate_code()



from keraspoj.src.backend.base import TikzOptions


class TikzStyle(TikzOptions):
    def __init__(self, *args, **kwargs):
        self.flags = args
        self.style_options = kwargs
        super().__init__(*args, **kwargs)

    def to_code(self):
        return f"{super().__str__()}"

    def __str__(self):
        return f"style with flags: {self.flags} and options: {self.style_options}"

    def __repr__(self):
        return str(self)


def create_node_style(color, node_distance, split_parts=10):
    return TikzStyle("rectangle split",
                     "rectangle split ignore empty parts",
                     "very thick",
                     **{
                         "rectangle split parts": split_parts,
                         "draw": f"{color}!60",
                         "fill": f"{color}!5",
                         "minimum width": "{width(\"Batch Normalisation\") + 8pt}",
                         "node distance": f"{node_distance}cm",
                         "outer sep": "0cm"
                     })



from keraspoj.src.backend.base import TikzElement
from keraspoj.src.util.tools import generate_uuid


class Grid(TikzElement):
    def __init__(self, to_x, to_y, from_x=0, from_y=0, grid_style_name="grid"):
        super().__init__("grid_" + generate_uuid())
        self.from_x = from_x
        self.from_y = from_y
        self.to_x = to_x
        self.to_y = to_y
        self.grid_style_name = grid_style_name

    def draw(self):
        return self.to_code()

    def to_code(self):
        # \draw[step=20pt, black, very thin] (0,0) grid (60,60);
        return rf"\draw[{self.grid_style_name}] ({self.from_x}, {self.from_y}) grid ({self.to_x}, {self.to_y});" + "\n"



from keraspoj.src.backend.base import TikzElement, TikzOptions
from keraspoj.src.util.tools import latexify


class Edge(TikzElement):
    def __init__(self, from_node, to_node, label="",
                 color="black",
                 edge_style="defaultEdge", label_style="defaultLabel",
                 extra_args: TikzOptions = TikzOptions()):
        self.from_node = from_node
        self.to_node = to_node
        self.label = label
        self.edge_style = edge_style
        self.label_style = label_style
        self.extra_style = extra_args
        self.color = color
        # super().__init__(f"edge_{from_node}_{to_node}", depends_on=[from_node, to_node])
        super().__init__(f"edge_{from_node}_{to_node}", depends_on=[])

    def draw(self):
        return self.to_code()

    def to_code(self):
        from_node = latexify(self.from_node)
        to_node = latexify(self.to_node)

        out = f"% edge from {from_node} to {to_node}\n"

        self.extra_style.add_option("draw", self.color)
        if self.color == 'black':
            self.extra_style.add_option('line width', '1pt')
        else:
            self.extra_style.add_option('line width', '1.6pt')
        style_str = f"-Stealth, {self.edge_style}"
        if self.extra_style is not None:
            style_str += f", {self.extra_style.to_code()}"
        out += rf"\draw[{style_str}] ({from_node}) " \
               rf"to node [{self.label_style}] {{{self.label}}} ({to_node});"
        return out + "\n"

    def __str__(self):
        return f"edge from {self.from_node} to {self.to_node} with label {self.label}"



from distutils.core import setup

import setuptools

setup(
    name='keraspoj',  # How you named your package folder (MyLib)
    packages=setuptools.find_packages(where='keraspoj'),
    version='0.3.2',  # Start with a small number and increase it with every change you make
    license='MIT',  # Chose a license from here: https://help.github.com/articles/licensing-a-repository
    description='yet another keras visualisation library',  # Give a short description about your library
    author='Jari Van Melckebeke',  # Type in your name
    author_email='jarivanmelckebeke@gmail.com',  # Type in your E-Mail
    url='https://github.com/jvanmelckebeke/model-visualizer',
    # Provide either the link to your github or to your website
    download_url='https://github.com/jvanmelckebeke/model-visualizer/archive/refs/tags/prerelease-2.tar.gz',
    # I explain this later on
    keywords=['tensorflow', 'keras', 'visualisation'],  # Keywords that define your package best
    install_requires=[  # I get to this in a second
        'numpy',
        'pandas',
        'keras',
        'matplotlib',
        'inflect',
        'jsonschema',
        'pyYAML',
        'Wand',
        'pygraphviz',
        'pydot',
        'pytz',
        'seaborn',
        'networkx'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Intended Audience :: Developers',  # Define that your audience are developers
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',  # Again, pick a license
        'Programming Language :: Python :: 3',  # Specify which pyhton versions that you want to support
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    package_dir={"": "keraspoj"},
    python_requires=">=3.6"
)



import os
import uuid

INDENT_SIZE = 4
INDENT_STR = ' ' * INDENT_SIZE


def str_shape(shape):
    if isinstance(shape, tuple):
        if shape[0] is None:
            return str_shape(shape[1:])
        if len(shape) == 1:
            return str(shape[0])
        return str(shape)
    elif isinstance(shape, list):
        if len(shape) == 1:
            return str_shape(shape[0])
        return str([str_shape(s) for s in shape])
    else:
        return str(shape)


def latexify(text):
    text = text.replace("_", r"-")
    text = text.replace(".", r"-")
    return text


def generate_uuid():
    return str(uuid.uuid4())[:8]


def run_command(cmd):
    exit_code = os.system(cmd)
    if exit_code != 0:
        raise Exception("Error while running command: " + cmd)


def pt_to_mm(pt):
    return pt / 2.834645669


def pretty_print(d, indent=0):
    for key, value in d.items():
        print('\t' * indent + "\"" + str(key) + "\":")
        if isinstance(value, dict):
            pretty_print(value, indent + 1)
        else:
            print('\t' * (indent + 1) + str(value))



from keraspoj.src.backend.node_group import NodeGroup
from keraspoj.src.backend.misc.position import Position
from keraspoj.src.diagram.layers.layer import Layer


class LayerGroup:
    def __init__(self, primary_layer: Layer, layer_before: Layer = None, layer_after: Layer = None, name: str = None):
        if name is None:
            name = f"{primary_layer.name}_group"
        self.name = name
        self.primary_layer = primary_layer
        self.layer_before = layer_before
        self.layer_after = layer_after

        self.layers = [primary_layer, layer_before, layer_after]
        self.layers = [l for l in self.layers if l is not None]

    def create_node_group(self, position: Position = None):
        return NodeGroup(self.primary_layer.create_node(position=position),
                         node_before=self.layer_before.create_node() if self.layer_before is not None else None,
                         node_after=self.layer_after.create_node() if self.layer_after is not None else None)

    @property
    def top_layer_name(self):
        if self.layer_before is not None:
            return self.layer_before.name
        return self.primary_layer.name

    @property
    def bottom_layer_name(self):
        if self.layer_after is not None:
            return self.layer_after.name
        return self.primary_layer.name

    @property
    def input_layers(self) -> list[str]:
        if self.layer_before is None:
            return [l.name for l in self.primary_layer.keras_input_layers]
        return [l.name for l in self.layer_before.keras_input_layers]

    @property
    def output_layers(self) -> list[str]:
        if self.layer_after is None:
            return [l.name for l in self.primary_layer.keras_output_layers]
        return [l.name for l in self.layer_after.keras_output_layers]

    def contains_layer_name(self, name):
        return any([l.name == name for l in self.layers])

    def __repr__(self):
        return f"LayerGroup({self.name}) primary layer {self.primary_layer.name} with before: " \
               f"{self.layer_before.name if self.layer_before is not None else None}, with after: " \
               f"{self.layer_after.name if self.layer_after is not None else None})"

    def __str__(self):
        return self.__repr__()



import math


class Position(dict):
    def __init__(self, x: float = 0, y: float = 0):
        super().__init__()
        self.x = round(x, 2)
        self.y = round(y, 2)

    def y_distance(self, other):
        return abs(self.y - other.y)

    def x_distance(self, other):
        return abs(self.x - other.x)

    def distance(self, other):
        return math.sqrt((self.x - other.x) ** 2 + (self.y - other.y) ** 2)

    def to_code(self):
        return f"at ({self.x}, {self.y})"



from keraspoj.src.backend.base import TikzElement
from keraspoj.src.backend.misc.position import Position
from keraspoj.src.backend.node import Node

DEBUG_OFFSET = 0

class NodeGroup(TikzElement):

    def __init__(self, primary_node: Node, node_before: Node = None, node_after: Node = None, name=None):
        if name is None:
            name = primary_node.name + "_group"
        self.name = name
        self.primary_node = primary_node
        self.node_before = node_before
        self.node_after = node_after

        super().__init__(name)

    @property
    def height(self):
        total_height = self.primary_node.height
        if self.node_before is not None:
            total_height += self.node_before.height
        if self.node_after is not None:
            total_height += self.node_after.height
        return total_height

    def draw(self):
        return self.to_code()

    def reposition_before_node(self):
        if self.node_before is None:
            return
        if self.node_after is None:
            self.node_before.position = Position(self.primary_node.position.x - DEBUG_OFFSET,
                                                 self.primary_node.position.y + self.height / 2)
        else:
            self.node_before.position = Position(self.primary_node.position.x - DEBUG_OFFSET,
                                                 (self.primary_node.position.y
                                                  + self.height / 2 - self.node_after.height / 2))

    def reposition_after_node(self):
        if self.node_after is None:
            return
        if self.node_before is None:
            self.node_after.position = Position(self.primary_node.position.x + DEBUG_OFFSET,
                                                self.primary_node.position.y - self.height / 2)
        else:
            self.node_after.position = Position(self.primary_node.position.x + DEBUG_OFFSET,
                                                (self.primary_node.position.y
                                                 - self.height / 2 + self.node_before.height / 2))

    def reposition_nodes(self):
        if self.node_before is not None:
            self.reposition_before_node()
        if self.node_after is not None:
            self.reposition_after_node()

    def to_code(self):
        code = f"% node group: {self.name}\n"
        self.reposition_nodes()
        # print(self)
        if self.node_before is not None:
            code += self.node_before.draw()

        if self.node_after is not None:
            code += self.node_after.draw()

        code += self.primary_node.draw()

        code += f"% end of node group: {self.name}\n"
        return code

    def __repr__(self):
        return str(self)

    def __str__(self):
        text = f"NodeGroup: {self.name} with primary node {self.primary_node.name}"
        if self.node_before is not None:
            text += f" and node before {self.node_before.name}"
        if self.node_after is not None:
            text += f" and node after {self.node_after.name}"
        return text
